#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
from __future__ import annotations
from abc import  abstractmethod, ABC
from typing import List

from soup_files import File, Directory
from convert_stream.models.m_imgelib import ABCImageObject, LibraryImage
from convert_stream.models.m_pdf_page import ABCPagePdf, LibraryPDF
from convert_stream.models.m_pdf_document import ABCDocumentPdf, ABCPagePdf


class ABCConvertPdf(ABC):
    """
        Converte um documento PDF ou Página de PDF em imagem(s).
    """
    def __init__(self):
        super().__init__()
        
    @abstractmethod
    def from_page_bytes(self, page_bytes:bytes, dpi:int=200) -> ABCImageObject:
        """
            Converte os bytes de uma página PDF em objeto Imagem.
        """
        pass
    
    @abstractmethod
    def from_page_pdf(self, page:ABCPagePdf, dpi:int=200) -> ABCImageObject:
        """
            Converte o um objeto PageDocumentPdf() em objeto Imagem.
        """
        pass
    
    @abstractmethod
    def inner_images(self, page_bytes:bytes) -> List[ABCImageObject]:
        """
            Converte todas as imagens embutidas em uma página PDF para uma lista de objetos Imagem.
        """
        pass
                    
        
class ABCImageConvertPdf(ABC):
    """Convete uma imagem em uma página PDF."""
    def __init__(self, library_image:LibraryImage=LibraryImage.OPENCV):
        super().__init__()
        self.library_image:LibraryImage = library_image
        
    @abstractmethod
    def from_image_file(self, file:File) -> ABCPagePdf:
        """
            Converte um arquivo de imagem em página PDF.
        """
        pass
    
    @abstractmethod
    def from_image(self, img:ABCImageObject) -> ABCDocumentPdf:
        """Converte um objeto imagem em uma página PDF."""
        pass
    
    @abstractmethod
    def from_image_bytes(self, img_bytes:bytes) -> ABCPagePdf:
        """Converte os bytes de uma imagem e uma página PDF."""
        pass
    
    
class ABCDocumentStream(ABC):
    """
        Objeto para manipular a exportação de dados PDF    
    """
    def __init__(self, *, library_pdf=LibraryPDF.FITZ, library_image=LibraryImage.OPENCV):
        self.library_pdf:LibraryPDF = library_pdf
        self.liabrary_image:LibraryImage = library_image
        
    @abstractmethod
    def is_null(self) -> bool:
        pass
    
    @abstractmethod
    def clear(self):
        pass
        
    @abstractmethod
    def to_file_pdf(self, f:File):
        """
            Exporta os dados para um arquivo PDF no disco.
        """
        pass
        
    @abstractmethod
    def to_files_pdf(self, outdir:Directory, *, prefix='pag'):
        """Exporta cada página para um novo arquivo PDF no disco"""
        pass
         
    @abstractmethod   
    def to_files_image(self, d:Directory, *, prefix:str='página-para-imagem'):
        """
            Exporta cada página PDF como um arquivo de imagem.
        """
        pass
            
    @abstractmethod
    def to_images(self) -> List[ABCImageObject]:
        """Retorna uma lista de imagens, dos documentos adicionados."""
        pass
    
    @abstractmethod
    def to_document(self) -> ABCDocumentPdf:
        """Converte os itens adicionados em documento"""
        pass
    
    def inner_images(self) -> List[ABCImageObject]:
        """
            Retorna uma lista com todas as imagens presentes em todas as páginas
        """
        pass
    
    @abstractmethod
    def inner_images_to_files(self, d:Directory, prefix:str='página-para-imagens') -> None:
        """
            Salva todas as imagens presentes nas páginas em arquivos de imagem.
        """
        pass
    
    @abstractmethod
    def add_page(self, p:ABCPagePdf):
        """Adiciona uma página PDF"""
        pass
    
    @abstractmethod
    def add_pages(self, pages:List[ABCPagePdf]):
        pass
       
    @abstractmethod     
    def add_page_pdf_bytes(self, bt:bytes):
        """Adicona os bytes de uma página PDF."""
        pass
          
    @abstractmethod  
    def add_file_pdf(self, f:File):
        pass
        
    @abstractmethod
    def add_files_pdf(self, files:List[File]):
        pass
    
    @abstractmethod
    def add_document(self, doc:ABCDocumentPdf):
        pass
    
    @abstractmethod
    def add_image(self, image: ABCImageObject):
        pass
    
    @abstractmethod
    def set_paisagem(self):
        """Define todas as páginas como paisagem."""
        pass
       

      
    
    








